#ifndef WakeComH
#define WakeComH

#include <SysUtils.hpp>
#include <SyncObjs.hpp>
#include <Classes.hpp>

#define DEFAULT_BAUD_RATE  19200 //   , 
#define DEFAULT_TX_TIMEOUT   100 //timeout    , 
#define DEFAULT_RX_TIMEOUT   100 //timeout    , 

#define DEFAULT_RTS_STATE  RTS_CONTROL_TOGGLE  // RTS  
#define DEFAULT_DTR_STATE  DTR_CONTROL_DISABLE // DTR  

#define FRAME     250   //  Wake

#define FEND     0xC0   //Frame END
#define FESC     0xDB   //Frame ESCape
#define TFEND    0xDC   //Transposed Frame END
#define TFESC    0xDD   //Transposed Frame ESCape

#define CRC_INIT 0xDE   //Initial CRC value

#define ERR_NO   0x00   //no error
#define ERR_TX   0x01   //Rx/Tx error
#define ERR_BU   0x02   //device busy error
#define ERR_RE   0x03   //device not ready error
#define ERR_PA   0x04   //parameters value error
#define ERR_NR   0x05   //no replay
#define ERR_NC   0x06   //no carrier

#define CMD_NOP     0   // 
#define CMD_ERR     1   //  
#define CMD_ECHO    2   // 
#define CMD_INFO    3   //   
#define CMD_SETADDR 4   // 
#define CMD_GETADDR 5   // 

#define LO(x)    (x & 0xFF)
#define HI(x)    ((x >> 8) & 0xFF)
#define BYTE1(x) LO(x)
#define BYTE2(x) HI(x)
#define BYTE3(x) ((x >> 16) & 0xFF)
#define BYTE4(x) ((x >> 24) & 0xFF)
#define WORD(h, l) ((h << 8) | l)
#define DWORD(b3,b2,b1,b0) ((WORD(b3,b2) << 16) | WORD(b1,b0))

#define _ENTER_COMMAND(x) x->Enter();try{
#define _LEAVE_COMMAND(x) }__finally{x->Leave();}

//------------------    WAKE: -----------------------

class TWakeCom : public TObject
{
private:
  HANDLE hCom;         //handle 
  DCB dcb;             // DCB
  DCB dcbc;            //  DCB
  COMMTIMEOUTS ComTo;  // timeout-
  COMMTIMEOUTS ComToc; //  timeout-
  AnsiString FName;    //  
  int FNumber;         // 
  int FBaudRate;       // 
  bool Platform_NT;    //  
  WORD BaseAddress;    //  UART (not for NT systems)
  bool RtsToggle;      //  RTS toggle (not for NT systems)

  void __fastcall SetNumber(int number);     //  
  void __fastcall SetName(AnsiString name);  //  

  int FTxAddr;           // 
  int FTxCmd;            //  
  int FTxPtr;            //  
  int FRxPtr;            //  
  AnsiString FTxCmdName; //  
  AnsiString FDeviceName; // 
  void __fastcall RxByte(unsigned char &b); // 
  BYTE __fastcall PortIn(WORD Address); //  (not for NT systems)
  void __fastcall ByteStuff(unsigned char b, DWORD &bptr, unsigned char *buff);
  void __fastcall ByteDeStuff(unsigned char &b);
  void __fastcall DowCRC(unsigned char b, unsigned char &crc); // CRC
  void __fastcall RxFrame(unsigned char &Addr,
                          unsigned char &Cmd,
                          unsigned char &Nbt,
                          unsigned char *Data); // 
  void __fastcall TxFrame(unsigned char Addr,
                          unsigned char Cmd,
                          unsigned char Nbt,
                          unsigned char *Data); // 
  unsigned char FTxData[FRAME]; // 
  unsigned char FRxData[FRAME]; // 

protected:
  TCriticalSection *CmdCrSec; //    
  void __fastcall NewCmd(int cmd, AnsiString name); //  
  void __fastcall AddByte(unsigned char b);  //   
  void __fastcall AddWord(unsigned short b); //   
  void __fastcall AddDWord(unsigned int b);  //    
  void __fastcall AddData(char *d, int count); //   
  void __fastcall SendCmd(int a); //    
  unsigned char __fastcall  GetByte(void);   //   
  unsigned short __fastcall  GetWord(void);  //   
  unsigned int __fastcall  GetDWord(void);   //    
  void __fastcall GetData(char *d, int count); //   

public:
  __fastcall TWakeCom();  //
  __fastcall ~TWakeCom(); //

  int __fastcall NameToNumber(AnsiString name); //   
  AnsiString __fastcall NumberToName(int number); //   
  void __fastcall GetPortList(TStringList *APortBox); //  
  bool __fastcall Access(int PortNumber); //    
  bool __fastcall Access(AnsiString PortName); //    
  void __fastcall Open(int PortNumber); //   
  void __fastcall Open(AnsiString PortName); //   
  void __fastcall GetDeviceList(TStringList *ADeviceBox); //  .
  void __fastcall GetNetDeviceList(AnsiString PortName,      // 
                                   TStringList *ADeviceBox); // .
  void __fastcall OpenDevice(AnsiString DevName); //    .
  void __fastcall Close(); // 
  void __fastcall SetBaudRate(int baudrate); //  
  void __fastcall SetMode(DWORD RtsMode, DWORD DtrMode); // 
  void __fastcall SetTo(DWORD RxTo, DWORD TxTo); // 
  void __fastcall Purge(void); //  

  __property int Number = {read=FNumber, write=SetNumber}; // 
  __property AnsiString PortName = {read=FName, write=SetName}; // 
  __property int BaudRate = {read=FBaudRate, write=SetBaudRate}; //
  __property AnsiString DeviceName = {read=FDeviceName}; // 

  AnsiString __fastcall GetInfo(int a); //   
};

//---------------------------------------------------------------------------

#endif
